# ------------------------------------------------------------------------------
#  File: local_config.py
#  Author: Jan Kukacka
#  Date: 7/2018
# ------------------------------------------------------------------------------
#  Module for local configurations.
#
#  Reads constants from file %HOME%/.work_config
# ------------------------------------------------------------------------------

from pathlib import Path
import json

_local_config_cache__ = None


def local(key, default=None):
    global _local_config_cache__

    if _local_config_cache__ is None:
        config_path = Path.home() / '.work_config'
        if config_path.exists():
            with config_path.open() as file:
                _local_config_cache__ = json.load(file)
        else:
            _local_config_cache__ = {}

    if key in _local_config_cache__:
        return _local_config_cache__[key]

    if default is not None:
        _local_config_cache__[key] = default
        return default

    return None



# ------------------------------------------------------------------------------
#  File: misc.py
#  Author: Jan Kukacka
#  Date: 6/2018
# ------------------------------------------------------------------------------
#  Miscelaneous functions
# ------------------------------------------------------------------------------


def ensure_list(var):
    '''
    Helping function for scenarios when variable (usually function parameter)
    is allowed to be an object or a list of objects.
    This returns canonical representation where even single objects are converted
    to a list containing a single object.

    Taken from https://stackoverflow.com/a/1416677/2042751.
    '''
    if isinstance(var, str):
        var = [var]
    ## Dictionaries are iterables but we consider them normal objects for this
    ## purpose.
    elif isinstance(var, dict):
        var = [var]
    else:
        try:
            iter(var)
        except TypeError:
            var = [var]
    return var



# ------------------------------------------------------------------------------
#  File: io.py
#  Author: Jan Kukacka
#  Date: 11/2018
# ------------------------------------------------------------------------------
#  Common functions for input/output operations
# ------------------------------------------------------------------------------

from copy import deepcopy
from pathlib import Path
import os

## For cached file access
__file_cache = {}


def load(filename, **kwargs):
    '''
    Loads file. Handles multiple formats:
        - .mat
        - .nii / .nii.gz
        - .mha
        - .hdr
        - .npy / .npz
        - .pkl
        - .json
        - .png
        - .jpeg / .jpg
        - .tiff / .tif
        - .bmp
        - .gif
        - dicom series (folder with .dcm)
        - .txt
    and selects the appropriate method to open them.

    # Arguments
    - filename: string with the file path or file_info dictionary containing
        the "filename" record.
    - kwargs:
        - key: string. Key in .mat or .npz file.
        - backend: string. For image files, may specify backend library (PIL or imageio)
    '''
    ## Handle cases when file_info is given
    if isinstance(filename, dict):
        return load(**filename)

    if not Path(filename).exists():
        raise FileNotFoundError(str(filename))
        # raise Exception('File not found')

    filename_str = str(filename).lower()

    if filename_str.endswith('.mat'):
        return load_matlab(filename, **kwargs)

    if (filename_str.endswith('.nii.gz')
        or filename_str.endswith('.nii')
        or filename_str.endswith('.mha')
        or filename_str.endswith('.hdr')):
        return load_nifti(filename)

    if filename_str.endswith('.npz') or filename_str.endswith('.npy'):
        return load_numpy(filename, **kwargs)

    if filename_str.endswith('.pkl'):
        return load_pickle(filename)

    if filename_str.endswith('.json'):
        return load_json(filename)

    if filename_str.endswith('.txt'):
        return load_txt(filename)

    if (filename_str.endswith('.png')
        or filename_str.endswith('.jpeg')
        or filename_str.endswith('.jpg')
        or filename_str.endswith('.tif')
        or filename_str.endswith('.tiff')
        or filename_str.endswith('.bmp')
        or filename_str.endswith('.gif')):
        return load_image(filename, **kwargs)

    if filename_str.endswith('.csv'):
        return load_csv(filename, **kwargs)

    if filename_str.endswith('.nrrd'):
        return load_nrrd(filename, **kwargs)

    ## DICOMs: filename is a name of a folder that contains some .dcm files
    if (os.path.isdir(filename)
        and any(file.endswith('.dcm') for file in os.listdir(filename))):
        return load_dicom_series(filename)

    raise Exception('Could not read the file')


def save(filename, data, overwrite=False, parents=False, **kwargs):
    '''
    Saves data to a file. Supports various file formats:
        - Nifti: .nii.gz
        - gif: .gif
        - pickle: .pkl
        - nrrd: .nrrd
        - numpy: .npy or .npz
        - json: .json
        - png: .png
        - csv: .csv
        - txt: .txt

    # Arguments:
        - filename: string with the file path or file_info dictionary containing
            the "filename" record.
        - data: array or dict or whatever suitable data to save.
        - overwrite: bool. If the file exists, is it okay to overwrite it?
            Default False.
        - parents: bool. Should the parent directories be created if they don't
            already exist? Default False.
        - kwargs:
            - is_vector: for nifti images. See save_nifti for details.
            - spacing: for nifti images. See save_nifti for details.
            - key: string. Key in .mat or .npz file.
    '''
    ## file_info dict
    if isinstance(filename, dict):
        filename = filename['filename']

    filename = Path(filename)

    ## Check if file exists
    ## Note: not safe to race conditions. If overwrite is False but some other
    ##       process creates the file after this check, it will get overwritten.
    if filename.exists() and not overwrite:
        try:
            import os
            raise OSError(os.errno.EEXIST, 'File exists. Use overwrite=True to overwrite it.')
        except AttributeError:
            import errno
            raise OSError(errno.EEXIST, 'File exists. Use overwrite=True to overwrite it.')

    ## Check if directory exists
    if not filename.parent.exists() and parents:
        filename.parent.mkdir(parents=True, exist_ok=True)

    if filename.name.endswith('.nii.gz'):
        save_nifti(filename, data, **kwargs)
    elif filename.name.endswith('.npy'):
        save_numpy(filename, data)
    elif filename.name.endswith('.npz'):
        save_numpy_archive(filename, data)
    elif filename.name.endswith('.nrrd'):
        save_nrrd(filename, data)
    elif filename.name.endswith('.pkl'):
        save_pickle(filename, data)
    elif filename.name.endswith('.gif'):
        save_gif(filename, data, **kwargs)
    elif filename.name.endswith('.json'):
        save_json(filename, data, **kwargs)
    elif filename.name.endswith('.png'):
        save_png(filename, data, **kwargs)
    elif filename.name.endswith('.csv'):
        save_csv(filename, data, **kwargs)
    elif filename.name.endswith('.txt'):
        save_txt(filename, data)
    else:
        raise Exception('Unsupported filetype.')


def load_file_cached_local(filename, copy=True, **kwargs):
    '''
    Works just as load(), but uses internal cache to avoid re-loading the same
    file multiple times. There is no cache size control in place, so use with
    caution.
    '''
    global __file_cache
    ## Handle file_info dictionaries
    if type(filename) == dict:
        key = filename['filename']
    else:
        key = filename
    if key not in __file_cache:
        __file_cache[key] = load(filename, **kwargs)
    if not copy:
        return __file_cache[key]
    else:
        return deepcopy(__file_cache[key])


def load_pickle(filename):
    '''
    Loads pickle file by its filename.
    '''
    ## Pickle files
    import pickle
    with open(filename, 'rb') as file:
        return pickle.load(file)


def save_pickle(filename, data):
    '''
    Save pickle file by its filename.
    '''
    ## Pickle files
    import pickle
    with open(filename, 'wb') as file:
        return pickle.dump(data, file, protocol=4)


def load_nifti(filename):
    '''
    Loads nifti file by its filename.
    '''
    ## Nifti files
    ## TODO: include spacing / resolution information
    import SimpleITK as sitk
    img = sitk.ReadImage(str(filename))
    return sitk.GetArrayFromImage(img)


def save_nifti(filename, data, is_vector=False, spacing=None, direction=None,
               origin=None):
    '''
    Saves "data" into nifti file "filename".

    # Arguments:
        - filename: string or Path with the full path to the output file.
        - data: 2D or 3D array with the image to save.
        - is_vector: bool. indicates if it is 3D image (False) or 2D
            multichannel image (True).
        - spacing: (optional) int or iterable of ints giving voxel spacing in
            the image in milimeters.
        - direction: (optional) information about the dimension ordering.
            e.g. 9-tuple of zeros and ones
        - origin: (optional) information about the image coord origin.
    '''
    ## Nifti files
    import SimpleITK as sitk
    img = sitk.GetImageFromArray(data, isVector=is_vector)

    if spacing is not None:
        ndims = data.ndim
        if is_vector:
            ndims -= 1
        if not hasattr(spacing, '__iter__'):
            spacing = [spacing] * ndims
        img.SetSpacing(spacing)

    if direction is not None:
        img.SetDirection(direction)

    if origin is not None:
        img.SetOrigin(origin)

    sitk.WriteImage(img, str(filename))


def load_numpy(filename, **kwargs):
    '''
    Loads numpy file by its filename and key (if it is a npz archive).
    '''
    import numpy as np
    file = np.load(filename)

    try:
        keys = file.keys()
        if 'key' not in kwargs:
            ## Is there a single key?
            if len(keys) == 1:
                return file[keys[0]]
            else:
                return file
        else:
            return file[kwargs['key']]
    except AttributeError:
        ## npy file has no keys
        return file


def save_numpy(filename, data):
    '''
    Saves a numpy array to a npy file
    '''
    import numpy as np
    np.save(filename, data)


def save_numpy_archive(filename, data):
    '''
    Saves a dict of arrays to a npz archive
    '''
    import numpy as np
    np.savez_compressed(filename, **data)


def load_matlab(filename, **kwargs):
    '''
    Loads matlab file by its filename and possibly key
    '''
    ## Matlab files
    import scipy.io
    import numpy as np
    try:
        file = scipy.io.loadmat(filename)
    except NotImplementedError:
        ## Thrown for h5py mat files
        import h5py
        file = h5py.File(filename,'r')

    if 'key' not in kwargs:
        ## Try automatically estimate the relevant key
        ## (the only one not starting with "__")
        keys = file.keys()
        keys = list(filter(lambda x: not x.startswith('__'), keys))
        if len(keys) == 1:
            ## Convert hdf files to ndarray but avoid unnecessary copy
            if isinstance(file[keys[0]], np.ndarray):
                return file[keys[0]]
            else:
                return np.array(file[keys[0]])
        else:
            return file
    else:
        ## Convert hdf files to ndarray but avoid unnecessary copy
        if isinstance(file[kwargs['key']], np.ndarray):
            return file[kwargs['key']]
        else:
            return np.array(file[kwargs['key']])


def load_json(filename):
    '''
    Loads json files
    '''
    import json
    with open(filename, 'r') as file:
        return json.load(file)


def save_json(filename, data):
    '''
    Saves json files
    '''
    import json
    with open(filename, 'w') as file:
        json.dump(data, file)


def load_image(filename, backend=None):
    '''
    Loads image, using a specific backend if needed

    # Arguments:
        - backend: 'PIL' or 'imageio'
    '''
    import numpy as np

    def _load_imageio(filename):
        import imageio
        return imageio.imread(filename)
    def _load_pil(filename):
        from PIL import Image
        return np.array(Image.open(filename))

    backends =  {'imageio': _load_imageio, 'PIL': _load_pil}

    if backend is not None:
        return backends[backend](filename)

    else:
        for backend in backends.values():
            try:
                return backend(filename)
            except ImportError:
                ## If some backend is not installed, try falling back to another
                pass
    raise Error('No suitable backend found. Install PIL or imageio to read images.')


def load_png(filename):
    '''
    Loads png files
    '''
    import imageio
    return imageio.imread(filename)


def save_png(filename, data, **kwargs):
    '''
    Saves png files
    '''
    import imageio
    return imageio.imwrite(filename, data, **kwargs)


def save_gif(filename, data, **kwargs):
    '''
    Saves gif image.

    # Arguments:
        - data: iterable with frames. Each frame should be of shape (height, width),
            (h,w,3) or (h,w,4).
        - filename: Output filename (full path). Should end with '.gif'
    '''
    import imageio
    imageio.mimsave(filename, data, **kwargs)


def load_dicom_series(folder_path):
    '''
    Loads a series of dicom images from a folder specified by folder_path.
    Based on https://simpleitk.readthedocs.io/en/master/Examples/DicomSeriesReader/Documentation.html

    # Arguments:
        - folder_path: Path to the folder with .dcm files to load.

    # Returns:
        - np array with image series
    '''
    import SimpleITK as sitk
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(folder_path)
    reader.SetFileNames(dicom_names)
    image = reader.Execute()
    return sitk.GetArrayFromImage(image)

def ensure_folder(path):
    '''Make sure a folder exists. Returns False if it cannot be created.'''
    import os
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
        except OSError as e:
            # print "Couldn't create output directory."
            return False
    return True

def load_csv(filename, **kwargs):
    '''
    Loads csv files.

    kwargs can use keyword 'pandas'. If True, Pandas module will be used.
    '''
    if 'pandas' in kwargs and kwargs['pandas']:
        import pandas
        return pandas.read_csv(filename)

    import csv
    data = []
    if 'delimiter' not in kwargs:
        kwargs['delimiter'] = ';'
    with open(filename, newline='') as csvfile:
        reader = csv.DictReader(csvfile, **kwargs)
        for row in reader:
            data.append(row)
    return data


def save_csv(filename, data, **kwargs):
    '''
    Saves csv files.

    If data is Pandas.DataFrame, DataFrame.to_csv() function is used.
    Otherwise, this functionality is not implemented yet.
    '''
    try:
        ## Try if it is pandas DataFrame
        data.to_csv(filename, **kwargs)
    except AttributeError as e:
        raise NotImplementedError('csv saving is only supported for Pandas data frames.')


def load_nrrd(filename, header=False):
    '''
    Loads nrrd files

    # Arguments:
        - header: bool. If True, returns also header dictionary as a second
            return item
    '''
    import nrrd
    data, file_header = nrrd.read(filename)
    if not header:
        return data
    else:
        return data, file_header


def save_nrrd(filename, data):
    '''
    Saves nrrd files
    '''
    import nrrd
    nrrd.write(str(filename), data)


def save_txt(filename, data):
    '''
    Saves txt files
    '''
    with open(filename, 'w') as file:
        file.write(data)

def load_txt(filename):
    '''
    Saves txt files
    '''
    with open(filename, 'r') as file:
        return file.read()



# ------------------------------------------------------------------------------
#  File: plots.py
#  Author: Jan Kukacka
#  Date: 7/2018
# ------------------------------------------------------------------------------
#  Plot utilities and standard plot functions
# ------------------------------------------------------------------------------

import numpy as np
import matplotlib.colors as mc
import matplotlib.pyplot as plt
import matplotlib.cm
from scipy.interpolate import interp1d
from .misc import ensure_list


__transparent = '#00000000'


def hide_ticks(axes=None, axis='both'):
    '''
    Hide ticks and their labels.

    # Arguments:
        - axes: axes object to hide ticks on. plt.gca() by default.
        - axis: 'x', 'y', 'both' (default). Which axis will be affected.
    '''
    if axes is None:
        axes = plt.gca()

    if axis == 'both' or axis == 'y':
        axes.set_yticklabels([])
        axes.yaxis.set_tick_params(length=0)
    if axis == 'both' or axis == 'x':
        axes.set_xticklabels([])
        axes.xaxis.set_tick_params(length=0)


def set_spine_thickness(axes=None, thickness=1):
    '''
    Allows setting thickness of spines of given axes

    # Arguments:
        - axes: axes object. Default is plt.gca()
        - thickness: positive float. Desired thicnkess.
    '''
    if axes is None:
        axes = plt.gca()

    for spine in axes.spines.values():
        spine.set_linewidth(thickness)


def scalebar(axes, res, ax_shape, len_cm=.5, color='w'):
    '''
    Displays scalebar on the plot.

    # Arguments:
        - axes: axes object to display scalebar on.
        - res: image resolution in horizontal dimension in microns
        - ax_shape: (height, width) of the displayed image
        - len_cm: float. Lenght of the scalebar in centimeters
        - color: valid matplotlib color. Color of the scalebar.
    '''
    len_pixels = len_cm * 10000 / res
    left = 0.05 * ax_shape[1]
    height_cm = .15
    height_in = height_cm * 0.393701
    ## Compute height and left-bottom point in display coordinates
    dpi_bottom = axes.transAxes.transform([0.05,0.95])
    dpi_height = axes.figure.dpi_scale_trans.transform([height_in]*2)
    # print(dpi_bottom)
    # print(dpi_height)

    ## Compute new bottom as bottom - height (shifted by height up)
    bottom = axes.transData.inverted().transform(dpi_bottom - dpi_height)
    # print(bottom)
    ## Since transData considers origin in lower left corner, we have to
    ## compute it like this:
    bottom = ax_shape[0] - bottom[1]
    ## Infer height
    height_pixels = 0.95 * ax_shape[0] - bottom
    # print(height_pixels)
    axes.add_artist(plt.Rectangle([left,bottom], len_pixels, height_pixels, color=color))


def cmap(*keypoints):
    '''
    Returns a linear segmented colormap from given keypoints.

    # Arguments
        - keypoints: Array-like of keypoints which can be either a color or a
            2-tuple of (index, color), where index is in range [0;1] and
            provides the location of the keypoint. If the index is not given,
            keypoints will be distributed equally on the range. `color` is any
            color format accepted by matplotlib, and `'t'` can be used for
            transparent color.

    # Returns:
        - colormap

    # Examples:
        Colormap ranging from transparent to red
        >>> cmap = happy.plots.cmap(['t', 'r'])
        Colormap with positions specified and transparent upper end
        >>> cmap = happy.plots.cmap([(0, 'black'), (0.2, 'xkcd: green'),
                                     (254/255, 'r'), (1, 't')])
    '''
    ## For compatibility with old version where keypoints was not *args parameter
    if len(keypoints) == 1:
        keypoints = keypoints[0]
    ## *keypoints is a tuple but we need to allow assignment in the transparent
    ## color replacement
    keypoints = list(ensure_list(keypoints))

    ## Replace color 't' for transparent
    for i in range(len(keypoints)):
        if keypoints[i] == 't':
            keypoints[i] = __transparent
        else:
            try:
                if keypoints[i][1] == 't':
                    ## Replace whole keypoint since old keypoint may be
                    ## immutable (e.g. a tuple)
                    keypoints[i] = (keypoints[i][0], __transparent)
            except:
                pass
    colormap = mc.LinearSegmentedColormap.from_list('', colors=keypoints)
    return colormap


def cmap_clip(colors=None, clip_up=True, clip_down=True):
    '''
    Creates colormap with gradient of the two given colors and transparent color
    for max and min value. Ideal for overlays on images.

    # Arguments:
        - colors: Array-like with colors of length at least 2. If None, 'red'
            and 'blue' are used.
        - clip_up: Bool. If True, highest values will be clipped. Default True.
        - clip_down: Bool. If True, lowest values will be clipped. Default True.
    '''
    if colors is None:
        colors = ('red', 'blue')
    keypoints = []
    if clip_down:
        keypoints.append((0, __transparent))
        keypoints.append((1/256, colors[0]))
    else:
        keypoints.append((0, colors[0]))

    ## Intermediate colors
    intermediate_spacing = 256 / (len(colors)-1)
    for i in range(len(colors)-2):
        keypoints.append(((i+1)*intermediate_spacing / 256, colors[i+1]))

    if clip_up:
        keypoints.append((255/256, colors[-1]))
        keypoints.append((1, __transparent))
    else:
        keypoints.append((1, colors[-1]))
    return cmap(keypoints)


def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


def normalize_color_list(colors, list_len):
    '''
    Ensures that the variable "colors" contains a list of colors of length
    list_len. If colors is a single color, it is repeated. If it is a list
    of colors already, it is expanded to the right lenght by repeating the
    last element or trimmed if it is too long.
    Also ensures that if colors is a single color represented as a list, it
    will be correctly understood.

    # Arguments:
        - colors: Color or list of colors
        - list_len: positive int. Length of the color list to be produced.

    # Returns:
        - color_list: List of colors of lenght list_len.
    '''
    if mc.is_color_like(colors):
        colors = [colors]
    else:
        colors = ensure_list(colors)
    if len(colors) < list_len:
        ## Expand short list
        colors = colors + [colors[-1]]*(list_len-len(colors))
    if len(colors) > list_len:
        ## Truncate long list
        colors = colors[:list_len]
    return colors


def smooth_plot(axes, x, y=None, smoothing=5, **kwargs):
    '''
    Smoothed version of pyplot.plot(). Plots single line on the given axes.

    # Arguments:
        - axes: axes obejct.
        - x, (y): arrays of shape (n,) with x and y coordinates. If y is missing,
            x is considered y and range(0,n) is considered x.
        - smoothing: positive int. how many samples should be interpolated
            between each pair of points?
        - kwargs: passed to normal plot.

    # Returns:
        - list of Line objects
    '''
    length = x.size
    if y is None:
        y = x
        x = np.arange(length)
    x_new = np.linspace(x.min(),x.max(), (length-1)*smoothing+1, endpoint=True)
    y_new = interp1d(x, y, kind='quadratic')
    return axes.plot(x_new, y_new(x_new), marker='o', markevery=smoothing, **kwargs)


def smooth_fill_between(x, y1, y2=0, smoothing=5, axes=None, **kwargs):
    '''
    Smoothed version of pyplot.fill_between(). Plots single shaded region on the
    given axes.

    # Arguments:
        - axes: axes obejct.
        - x, (y1, y2): arrays of shape (n,) with x and y coordinates. If y2 is missing,
            it is considered to be zero
        - smoothing: positive int. how many samples should be interpolated
            between each pair of points?
        - axes: axes object to plot on. If None, plt.gca() is used.
        - kwargs: passed to normal plot.

    # Returns:
        - A PolyCollection with plotted polygons
    '''
    if axes is None:
        axes = plt.gca()
    length = x.size
    x_new = np.linspace(x.min(),x.max(), (length-1)*smoothing+1, endpoint=True)
    y1_new = interp1d(x, y1, kind='quadratic')
    y2_new = interp1d(x, y2, kind='quadratic')
    return axes.fill_between(x_new, y1_new(x_new), y2_new(x_new), **kwargs)


def imshow(axes, img, alpha, cmap=None, norm=None, **kwargs):
    '''
    Extending pyplot.imshow with ability to give alpha as an array

    # Arguments:
        - axes: axes object.
        - img: array of shape (height, width) or (height, width, channels) where
            channels is either 1, 3 or 4. If it is 4 (alpha is given too), then
            the original alpha and given alpha are multiplied.
        - alpha: array of shape (height, width) or scalar (behaves like normal
            imshow then)
        - cmap: colormap to use. Defaults to rcParams default.
        - norm: normalization function to use. Defaults to Normalize.
        - kwargs: passed to pyplot.imshow.
    '''
    ## TODO: Change signature to allow axes=None as in other functions
    ## Set defaults
    if cmap is None:
        ## returns default colormap
        cmap = matplotlib.cm.get_cmap()
    if norm is None:
        norm = mc.Normalize()

    ## Get RGBA representation of the input
    if img.ndim > 2 and img.shape[2] == 1:
        img = img[...,0]
    if img.ndim == 2:
        img_rgba = cmap(norm(img))
    if img.ndim == 3 and img.shape[2] == 3:
        ## Image is already given as colors, append alpha channels
        img_rgba = np.concatenate([img, np.ones(img.shape[:-1] + (1,))],
                                  axis=-1)
    if img.ndim == 3 and img.shape[2] == 4:
        ## Image is already in the correct format
        img_rgba = img

    ## Blend alpha
    img_rgba[...,-1] = img_rgba[...,-1] * alpha
    return axes.imshow(img_rgba, **kwargs)


def boxplot(data, ax=None, edge_color=None, fill_color=None, **kwargs):
    '''
    Extends the functionality of standard pyplot boxplot function by spcifying
    colors.
    Inspired by https://stackoverflow.com/questions/41997493/python-matplotlib-boxplot-color

    # Arguments:
        - data: array or sequence of vectors. (Same as original)
        - ax: axes object. If None, plt.gca() is used.
        - edge_color: Color or list of colors for lines and box borders
        - fill_color: Color or list of colors for filling the boxes
        - kwargs: Passed to pyplot.boxplot

    # Retuns:
        - dict of boxplot artists (same as original).
    '''
    if ax is None:
        ax = plt.gca()
    use_patch_artist = fill_color is not None
    bp = ax.boxplot(data, patch_artist=use_patch_artist, **kwargs)

    if edge_color is not None:
        edge_color = normalize_color_list(edge_color, len(bp['boxes']))
        ## Single elements
        for element_type in ['boxes', 'means', 'medians']:
            for element, color in zip(bp[element_type], edge_color):
                plt.setp(element, color=color)

        ## Paired elements
        for element_type in ['whiskers', 'caps']:
            for i, color in enumerate(edge_color):
                plt.setp(bp[element_type][2*i], color=color)
                plt.setp(bp[element_type][2*i+1], color=color)

        ## Fliers (use "markeredgecolor")
        for flier, color in zip(bp['fliers'], edge_color):
            plt.setp(flier, markeredgecolor=color)

    if fill_color is not None:
        fill_color = normalize_color_list(fill_color, len(bp['boxes']))
        for patch, color in zip(bp['boxes'], fill_color):
            patch.set(facecolor=color)
    return bp


def matlab_cmap():
    '''
    Returns matlab colormap
    '''
    return cmap([[0.2081,0.1663,0.5292],[0.2116238095,0.1897809524,0.5776761905],[0.212252381,0.2137714286,0.6269714286],[0.2081,0.2386,0.6770857143],[0.1959047619,0.2644571429,0.7279],[0.1707285714,0.2919380952,0.779247619],[0.1252714286,0.3242428571,0.8302714286],[0.0591333333,0.3598333333,0.8683333333],[0.0116952381,0.3875095238,0.8819571429],[0.0059571429,0.4086142857,0.8828428571],[0.0165142857,0.4266,0.8786333333],[0.032852381,0.4430428571,0.8719571429],[0.0498142857,0.4585714286,0.8640571429],[0.0629333333,0.4736904762,0.8554380952],[0.0722666667,0.4886666667,0.8467],[0.0779428571,0.5039857143,0.8383714286],[0.079347619,0.5200238095,0.8311809524],[0.0749428571,0.5375428571,0.8262714286],[0.0640571429,0.5569857143,0.8239571429],[0.0487714286,0.5772238095,0.8228285714],[0.0343428571,0.5965809524,0.819852381],[0.0265,0.6137,0.8135],[0.0238904762,0.6286619048,0.8037619048],[0.0230904762,0.6417857143,0.7912666667],[0.0227714286,0.6534857143,0.7767571429],[0.0266619048,0.6641952381,0.7607190476],[0.0383714286,0.6742714286,0.743552381],[0.0589714286,0.6837571429,0.7253857143],[0.0843,0.6928333333,0.7061666667],[0.1132952381,0.7015,0.6858571429],[0.1452714286,0.7097571429,0.6646285714],[0.1801333333,0.7176571429,0.6424333333],[0.2178285714,0.7250428571,0.6192619048],[0.2586428571,0.7317142857,0.5954285714],[0.3021714286,0.7376047619,0.5711857143],[0.3481666667,0.7424333333,0.5472666667],[0.3952571429,0.7459,0.5244428571],[0.4420095238,0.7480809524,0.5033142857],[0.4871238095,0.7490619048,0.4839761905],[0.5300285714,0.7491142857,0.4661142857],[0.5708571429,0.7485190476,0.4493904762],[0.609852381,0.7473142857,0.4336857143],[0.6473,0.7456,0.4188],[0.6834190476,0.7434761905,0.4044333333],[0.7184095238,0.7411333333,0.3904761905],[0.7524857143,0.7384,0.3768142857],[0.7858428571,0.7355666667,0.3632714286],[0.8185047619,0.7327333333,0.3497904762],[0.8506571429,0.7299,0.3360285714],[0.8824333333,0.7274333333,0.3217],[0.9139333333,0.7257857143,0.3062761905],[0.9449571429,0.7261142857,0.2886428571],[0.9738952381,0.7313952381,0.266647619],[0.9937714286,0.7454571429,0.240347619],[0.9990428571,0.7653142857,0.2164142857],[0.9955333333,0.7860571429,0.196652381],[0.988,0.8066,0.1793666667],[0.9788571429,0.8271428571,0.1633142857],[0.9697,0.8481380952,0.147452381],[0.9625857143,0.8705142857,0.1309],[0.9588714286,0.8949,0.1132428571],[0.9598238095,0.9218333333,0.0948380952],[0.9661,0.9514428571,0.0755333333],[0.9763,0.9831,0.0538]])



# ------------------------------------------------------------------------------
#  File: ndarray.py
#  Author: Jan Kukacka
#  Date: 6/2018
# ------------------------------------------------------------------------------
#  Functions to help handling ndarrays
# ------------------------------------------------------------------------------

import numpy as np


def diag(array):
    '''
    Convert last axis of an array to a diagonal matrix, leave preceding
    dimensions unchanged.

    # Arguments:
        - array: numpy array of shape [..., N]

    # Returns:
        - array: numpy array of shape [..., N, N] with elements of the input
            array on the diagonal of the last 2 dimensions

    # Example:
    >>  diag(np.array([[1,2,3],
                       [4,5,6]]))
        [[[1, 0, 0],
          [0, 2, 0],
          [0, 0, 3]],
         [[4, 0, 0],
          [0, 5, 0],
          [0, 0, 6]]]
    '''
    array_flat = np.reshape(array, (-1, array.shape[-1]))
    result = np.empty(array_flat.shape + (array.shape[-1],))
    for i in range(array_flat.shape[0]):
        result[i] = np.diag(array_flat[i])
    return np.reshape(result, array.shape + (array.shape[-1],))


def collapse_diag(array):
    '''
    Convert matrix represented by the last 2 axes of an array to 1D array
    containing the diagonal of the matrix, leave preceding dimensions unchanged.

    # Arguments:
        - array: numpy array of shape [..., N, N]

    # Returns:
        - array: numpy array of shape [..., N] with elements of the diagonal of
            the last 2 dimensions in the last dimension

    # Example:
    >>  collapse_diag(np.array([[[1, 0, 0],
                                 [0, 2, 0],
                                 [0, 0, 3]],
                                [[4, 0, 0],
                                 [0, 5, 0],
                                 [0, 0, 6]]]))
         [[1,2,3],
          [4,5,6]]
    '''
    array_flat = np.reshape(array, (-1, array.shape[-1], array.shape[-1]))
    result = np.empty(array_flat.shape[:-1])
    for i in range(array_flat.shape[0]):
        result[i] = np.diag(array_flat[i])
    return np.reshape(result, array.shape[:-1])


def collapse_axes(a, preserve=None):
    '''
    Collapses (flattens) all the array except for axes specified in the parameter
    `preserve`, which will be kept. Dimensions between preserved axes will be
    collapsed.

    # Arguments:
        - a: input numpy array of shape [n1, n2, n3, ..., nk]
        - preserve: int or iterable of axes that should be preserved. Negative
            numbers may specify axes indexed from the end

    # Returns:
        - reshaped array with collapsed axes not specified in the preserve param

    # Example:
    >>  a = np.zeros((5,3,8,6,6,10))
    >>  collapse_axes(a, preserve=(1,-1)).shape
        (5,3,288,10)
    >>  collapse_axes(a, preserve=(3)).shape
        (320,6,60)
    '''
    if preserve is None:
        return a.ravel()

    from .misc import ensure_list
    preserve = ensure_list(preserve)

    n_dim = a.ndim
    shape = a.shape

    ## Convert preserved axes to positive integers
    preserve = [i if i >= 0 else n_dim+i for i in preserve]
    ## Sort dimensions
    preserve = sorted(list(set(preserve)))

    ## Compute new array dimensions
    new_shape = []
    accumulator = 1
    collapse = False
    for axis, size in enumerate(shape):
        ## Check if current axis should be preserved
        ## (it would be at the start of the current preserved list)
        if len(preserve) == 0 or axis < preserve[0]:
            ## When collapsing an axis, increase accumulator size and raise a flag
            accumulator = accumulator * size
            collapse = True
        else:
            ## This axis should be preserved
            if collapse:
                ## If there were some collapsed axes before this axis, add their
                ## accumulated shape to the new shape and reset flags
                new_shape.append(accumulator)
                accumulator = 1
                collapse = False
            ## Add preserved axes shape and remove it from the preserved list
            new_shape.append(size)
            del preserve[0]
    if collapse:
        ## Do not forget to append remaining axes at the end
        new_shape.append(accumulator)

    return a.reshape(new_shape)
